#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/mman.h>
#include "ports.h"

int   physical_memory_fd = -1;
void *ports_base;

void open_ports(u_int32_t base_addr, u_int32_t span) {
    if (physical_memory_fd < 0)
        physical_memory_fd = open("/dev/mem", O_RDWR | O_SYNC);
    if (physical_memory_fd < 0) {
        printf("Failed to open /dev/mem for read/write -- permissions perhaps?\n");
        exit(-__LINE__);
    }
    ports_base = mmap(NULL, span, PROT_READ | PROT_WRITE, MAP_SHARED, physical_memory_fd, base_addr);
    if (ports_base == MAP_FAILED) {
        printf("Failed to mmap physical memory page.\n");
        exit(-__LINE__);
    }
}

void close_ports() {
    if (physical_memory_fd >= 0)
        close(physical_memory_fd);
}

#define FIFO_FILL_LEVEL     0
#define FIFO_STATUS         4
#define FIFO_STATUS_FULL    0x01
#define FIFO_STATUS_EMPTY   0x02
#define FIFO_EVENT          8
#define FIFO_INTERRUPT_ENABLE   12
#define FIFO_ALMOST_FULL    16
#define FIFO_ALMOST_EMPTY   20

struct  fifo32 {
    volatile u_int32_t  *data_port;
    volatile u_int32_t  *fill_level_port;
    volatile u_int32_t  *status_port;
};

void fifo32_init(
    struct fifo32      *fifo,
    u_int32_t   data_address,
    u_int32_t   status_address) {
    fifo->data_port = (u_int32_t *) (data_address + (u_int32_t)ports_base);
    fifo->fill_level_port = (u_int32_t *) (status_address + FIFO_FILL_LEVEL + (u_int32_t)ports_base);
    fifo->status_port = (u_int32_t *) (status_address + FIFO_STATUS + (u_int32_t)ports_base);
}

int fifo32_empty(struct fifo32 *fifo) {
    return (*fifo->status_port & FIFO_STATUS_EMPTY) ? 1 : 0;
}

int fifo32_full(struct fifo32 *fifo) {
    return (*fifo->status_port & FIFO_STATUS_FULL) ? 1 : 0;
}

void fifo32_write(struct fifo32 *fifo, u_int32_t  v) {
    while (fifo32_full(fifo))
        ;
    *fifo->data_port = v;
}

u_int32_t fifo32_read(struct fifo32 *fifo) {
    while (fifo32_empty(fifo))
        ;
    return *fifo->data_port;
}

u_int32_t fifo32_entries(struct fifo32 *fifo) {
    return *fifo->fill_level_port;
}

#define CMD_ECHO        (0UL)
#define CMD_SET_ADDRESS (1UL)
#define CMD_WRITE       (2UL)
#define CMD_READ        (3UL)
#define CMD_FIELD_SHIFT 28
#define ADDRESS_WIDTH   25

u_int32_t   form_echo_cmd(u_int16_t    echo_data) {
    return (CMD_ECHO << CMD_FIELD_SHIFT) | (u_int32_t) echo_data;
}

u_int32_t   form_set_address_cmd(u_int32_t  address) {
    return (CMD_SET_ADDRESS << CMD_FIELD_SHIFT) | (address & ((1UL << ADDRESS_WIDTH) - 1));
}

u_int32_t   form_write_cmd(u_int64_t    write_data) {
    return (CMD_WRITE << CMD_FIELD_SHIFT) | (u_int32_t) write_data;
}

u_int32_t   form_read_cmd() {
    return (CMD_READ << CMD_FIELD_SHIFT);
}

void    write_data(struct fifo32 *output, u_int32_t    address, u_int16_t  data) {
    fifo32_write(output, form_set_address_cmd(address));
    fifo32_write(output, form_write_cmd(data));
}

u_int16_t   read_data(struct fifo32 *input, struct fifo32 *output, u_int32_t  address) {
    fifo32_write(output, form_set_address_cmd(address));
    fifo32_write(output, form_read_cmd());
    return (u_int16_t) fifo32_read(input);
}

int main(int argc, char *argv[]) {
    struct  fifo32  input, output;
    u_int32_t   i = 0, v, a, v1;

    open_ports(0xc0000000, 512);

    fifo32_init(&output, FIFO_0_IN_BASE, FIFO_0_IN_CSR_BASE);
    fifo32_init(&input, FIFO_1_OUT_BASE, FIFO_1_OUT_CSR_BASE);

    // drain the input fifo of any left over cruft
    while (!fifo32_empty(&input))
        fifo32_read(&input);

    if (argc != 1) {
        if (argv[1][0] == 'w') {
            a = atoi(argv[2]);
            v = atoi(argv[3]);
            write_data(&output, a, v);
            v1 = read_data(&input, &output, a);
            printf("Wrote %d to %d and read back %d\n", v, a, v1);
        } else
        if (argv[1][0] == 'r') {
            a = atoi(argv[2]);
            v1 = read_data(&input, &output, a);
            printf("Address %d read back %d\n", a, v1);
        } else
        if (argv[1][0] == 'e') {
            v = atoi(argv[2]);
            fifo32_write(&output, form_echo_cmd(v));
            v1 = fifo32_read(&input);
            printf("Echo wrote %d read back %d\n", v, v1);
        }
        close_ports();
        exit(0);
    }

    for (i = 0; i < 1024; i++) {
        fifo32_write(&output, form_echo_cmd(i));
        if ((v = fifo32_read(&input)) != i) {
            printf("Error: wrote %d but read back %d\n", i, v);
            exit(-__LINE__);
        }
    }

    // a more complex example
    for (i = 0; i < 16; i++)
        fifo32_write(&output, form_echo_cmd(i));

    for (i = 0; i < 16; i++)
        if ((v = fifo32_read(&input)) != i) {
            printf("Error: wrote %d but read back %d\n", i, v);
            exit(-__LINE__);
        }

    printf("Echo Test pasted!\n");

    printf("Writing SDRAM\n");
    for (i = 0; i < 1024; i++)
        write_data(&output, i, i*2);

    printf("Performing readback from SDRAM\n");
    for (i = 0; i < 1024; i++) {
        v = read_data(&input, &output, i);
        if (v != (i*2)) {
            printf("Error: expected address: %d to have value %d got %d\n", i, i*2, v);
        }
    }

    close_ports();
}


